# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2021 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
App Handler
"""

from __future__ import unicode_literals, absolute_import

# import re

import six
from sqlalchemy import orm

from rattail.util import load_object


class AppHandler(object):
    """
    Base class and default implementation for top-level Rattail app handler.

    aka. "the handler to handle all handlers"

    aka. "one handler to bind them all"
    """

    def __init__(self, config):
        self.config = config

    def get_board_handler(self, **kwargs):
        if not hasattr(self, 'board_handler'):
            from rattail.board import get_board_handler
            self.board_handler = get_board_handler(self.config, **kwargs)
        return self.board_handler

    def get_clientele_handler(self, **kwargs):
        if not hasattr(self, 'clientele_handler'):
            from rattail.clientele import get_clientele_handler
            self.clientele_handler = get_clientele_handler(self.config, **kwargs)
        return self.clientele_handler

    def get_employment_handler(self, **kwargs):
        if not hasattr(self, 'employment_handler'):
            from rattail.employment import get_employment_handler
            self.employment_handler = get_employment_handler(self.config, **kwargs)
        return self.employment_handler

    def get_feature_handler(self, **kwargs):
        if not hasattr(self, 'feature_handler'):
            from rattail.features import FeatureHandler
            self.feature_handler = FeatureHandler(self.config, **kwargs)
        return self.feature_handler

    def get_people_handler(self, **kwargs):
        if not hasattr(self, 'people_handler'):
            from rattail.people import get_people_handler
            self.people_handler = get_people_handler(self.config, **kwargs)
        return self.people_handler

    def get_products_handler(self, **kwargs):
        if not hasattr(self, 'products_handler'):
            from rattail.products import get_products_handler
            self.products_handler = get_products_handler(self.config, **kwargs)
        return self.products_handler

    def get_report_handler(self, **kwargs):
        if not hasattr(self, 'report_handler'):
            from rattail.reporting import get_report_handler
            self.report_handler = get_report_handler(self.config, **kwargs)
        return self.report_handler

    def get_session(self, obj):
        """
        Returns the SQLAlchemy session with which the given object is
        associated.  Simple convenience wrapper around
        ``sqlalchemy.orm.object_session()``.
        """
        return orm.object_session(obj)

    def make_session(self, **kwargs):
        """
        Creates and returns a new SQLAlchemy session for the Rattail DB.
        """
        from rattail.db import Session
        return Session(**kwargs)

    def phone_number_is_invalid(self, number):
        """
        This method should validate the given phone number string, and if the
        number is *not* considered valid, this method should return the reason
        as string.  If the number is valid, returns ``None``.
        """
        from rattail.db.util import normalize_phone_number, validate_phone_number

        # strip non-numeric chars, and make sure we have 10 left
        normal = normalize_phone_number(number)
        if len(normal) != 10:
            return "Phone number must have 10 digits"

        # run it through the "traditional" validation, from FormEncode
        try:
            number = validate_phone_number(number, error=True)
        except Exception as error:
            return six.text_type(error)


class GenericHandler(object):
    """
    Base class for misc. "generic" feature handlers.

    Most handlers which exist for sake of business logic, should inherit from
    this.
    """

    def __init__(self, config, **kwargs):
        self.config = config
        self.enum = self.config.get_enum()
        self.model = self.config.get_model()
        self.app = self.config.get_app()

    def get_session(self, obj):
        """
        Convenience wrapper around :meth:`AppHandler.get_session()`.
        """
        return self.app.get_session(obj)

    def make_session(self):
        """
        Convenience wrapper around :meth:`AppHandler.make_session()`.
        """
        return self.app.make_session()

    def cache_model(self, session, model, **kwargs):
        """
        Convenience method which invokes :func:`rattail.db.cache.cache_model()`
        with the given model and keyword arguments.
        """
        from rattail.db import cache
        return cache.cache_model(session, model, **kwargs)


def make_app(config, **kwargs):
    """
    Create and return the configured :class:`AppHandler` instance.
    """
    spec = config.get('rattail', 'app.handler')
    if spec:
        factory = load_object(spec)
    else:
        factory = AppHandler
    return factory(config, **kwargs)



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2021 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
People Handler
"""

from __future__ import unicode_literals, absolute_import

from rattail.util import load_object
from rattail.app import GenericHandler


class PeopleHandler(GenericHandler):
    """
    Base class and default implementation for people handlers.
    """

    def normalize_full_name(self, first, last, **kwargs):
        from rattail.db.util import normalize_full_name
        return normalize_full_name(first, last)

    def update_names(self, person, **kwargs):
        """
        Update name(s) for the given person.

        :param person: Reference to a ``Person`` record.
        :param first: First name for the person.
        :param middle: Middle name for the person.
        :param last: Last name for the person.
        :param full: Full (display) name for the person.
        """
        if 'first' in kwargs:
            person.first_name = kwargs['first']

        if 'middle' in kwargs:
            person.middle_name = kwargs['middle']

        if 'last' in kwargs:
            person.last_name = kwargs['last']

        if 'full' in kwargs:
            if kwargs['full']:
                person.display_name = kwargs['full']
            else:
                person.display_name = self.normalize_full_name(
                    person.first_name, person.last_name)
        elif 'first' in kwargs and 'last' in kwargs:
            person.display_name = self.normalize_full_name(
                person.first_name, person.last_name)


def get_people_handler(config, **kwargs):
    """
    Create and return the configured :class:`PeopleHandler` instance.
    """
    spec = config.get('rattail', 'people.handler')
    if spec:
        factory = load_object(spec)
    else:
        factory = PeopleHandler
    return factory(config, **kwargs)



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2019 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Handler for generic product batches
"""

from __future__ import unicode_literals, absolute_import

from sqlalchemy import orm

from rattail.db import model
from rattail.batch import BatchHandler, get_batch_handler


class ProductBatchHandler(BatchHandler):
    """
    Handler for generic product batches.
    """
    batch_model_class = model.ProductBatch

    def should_populate(self, batch):
        if batch.input_filename:
            return True
        return False

    def populate(self, batch, progress=None):
        if batch.input_filename:
            return self.populate_from_file(batch, progress=progress)

    def populate_from_file(self, batch, progress=None):
        raise NotImplementedError

    def refresh_row(self, row):
        if not row.product:
            if not row.item_entry:
                row.status_code = row.STATUS_MISSING_KEY
                return

            session = orm.object_session(row)
            row.product = self.locate_product_for_entry(session, row.item_entry)
            if not row.product:
                row.status_code = row.STATUS_PRODUCT_NOT_FOUND
                return

        product = row.product
        row.upc = product.upc
        row.item_id = product.item_id
        row.brand_name = product.brand.name if product.brand else None
        row.description = product.description
        row.size = product.size

        dept = product.department
        row.department = dept
        row.department_number = dept.number if dept else None
        row.department_name = dept.name if dept else None

        subdept = product.subdepartment
        row.subdepartment = subdept
        row.subdepartment_number = subdept.number if subdept else None
        row.subdepartment_name = subdept.name if subdept else None

        row.category = product.category
        row.family = product.family
        row.reportcode = product.report_code

        cost = product.cost
        row.vendor = cost.vendor if cost else None
        row.vendor_item_code = cost.code if cost else None
        row.regular_cost = cost.unit_cost if cost else None
        row.current_cost = cost.discount_cost if cost else None
        row.current_cost_starts = cost.discount_starts if row.current_cost else None
        row.current_cost_ends = cost.discount_ends if row.current_cost else None

        regprice = product.regular_price
        curprice = product.current_price
        sugprice = product.suggested_price
        row.regular_price = regprice.price if regprice else None
        row.current_price = curprice.price if curprice else None
        row.current_price_starts = curprice.starts if curprice else None
        row.current_price_ends = curprice.ends if curprice else None
        row.suggested_price = sugprice.price if sugprice else None

        row.status_code = row.STATUS_OK

    def execute(self, batch, user=None, action='make_label_batch', progress=None, **kwargs):

        if action == 'make_label_batch':
            result = self.make_label_batch(batch, user, progress=progress)

        elif action == 'make_pricing_batch':
            result = self.make_pricing_batch(batch, user, progress=progress)

        else:
            raise RuntimeError("Batch execution action is not supported: {}".format(action))

        return result

    def make_label_batch(self, product_batch, user, progress=None):
        handler = get_batch_handler(self.config, 'labels',
                                    default='rattail.batch.labels:LabelBatchHandler')
        session = orm.object_session(product_batch)
        label_batch = handler.make_batch(session, created_by=user,
                                         description=product_batch.description,
                                         notes=product_batch.notes)
        label_batch.product_batch = product_batch
        handler.do_populate(label_batch, user, progress=progress)
        return label_batch

    def make_pricing_batch(self, product_batch, user, progress=None):
        handler = get_batch_handler(self.config, 'pricing',
                                    default='rattail.batch.pricing:PricingBatchHandler')
        session = orm.object_session(product_batch)
        pricing_batch = handler.make_batch(session, created_by=user,
                                           description=product_batch.description,
                                           notes=product_batch.notes)
        pricing_batch.product_batch = product_batch
        handler.do_populate(pricing_batch, user, progress=progress)
        return pricing_batch



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2021 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Handler for "delete product" batches
"""

from __future__ import unicode_literals, absolute_import

from sqlalchemy import orm

from rattail.db import model
from rattail.batch import BatchHandler


class DeleteProductBatchHandler(BatchHandler):
    """
    Handler for delete product batches.
    """
    batch_model_class = model.DeleteProductBatch

    def should_populate(self, batch):
        if hasattr(batch, 'products'):
            return True
        return False

    def populate(self, batch, progress=None):
        if hasattr(batch, 'products'):
            return self.populate_from_query(batch, progress=progress)

    def populate_from_query(self, batch, progress=None):
        session = orm.object_session(batch)

        def append(product, i):
            row = self.make_row()
            row.item_entry = product.uuid
            row.product = product
            self.add_row(batch, row)
            if i % 200 == 0:
                session.flush()

        self.progress_loop(append, batch.products, progress,
                           message="Adding products to batch")

    def refresh_row(self, row):
        if not row.product:
            row.status_code = row.STATUS_PRODUCT_NOT_FOUND
            return

        self.refresh_product_basics(row)
        row.status_code = row.STATUS_OK
